from forepaas.dwh.connect import connect
from forepaas.worker.connector import bulk_insert
import logging
from time import sleep
from requests import get
from datetime import datetime as dt
from datetime import timedelta as td
from os import makedirs
import re
import pandas as pd
import numpy as np
# LZH形式のファイルを解凍するパッケージ lhafile をインポート
import lhafile

# URLの固定部分を指定
FIXED_URL = "http://www1.mbrace.or.jp/od2/B/"

CSV_FILE_HEADER = "レース日,レース場,レース回,\
1枠_艇番,1枠_登録番号,1枠_選手名,1枠_年齢,1枠_支部,1枠_体重,1枠_級別,\
1枠_全国勝率,1枠_全国2連対率,1枠_当地勝率,1枠_当地2連対率,\
1枠_モーター番号,1枠_モーター2連対率,1枠_ボート番号,1枠_ボート2連対率,\
2枠_艇番,2枠_登録番号,2枠_選手名,2枠_年齢,2枠_支部,2枠_体重,2枠_級別,\
2枠_全国勝率,2枠_全国2連対率,2枠_当地勝率,2枠_当地2連対率,\
2枠_モーター番号,2枠_モーター2連対率,2枠_ボート番号,2枠_ボート2連対率,\
3枠_艇番,3枠_登録番号,3枠_選手名,3枠_年齢,3枠_支部,3枠_体重,3枠_級別,\
3枠_全国勝率,3枠_全国2連対率,3枠_当地勝率,3枠_当地2連対率,\
3枠_モーター番号,3枠_モーター2連対率,3枠_ボート番号,3枠_ボート2連対率,\
4枠_艇番,4枠_登録番号,4枠_選手名,4枠_年齢,4枠_支部,4枠_体重,4枠_級別,\
4枠_全国勝率,4枠_全国2連対率,4枠_当地勝率,4枠_当地2連対率,\
4枠_モーター番号,4枠_モーター2連対率,4枠_ボート番号,4枠_ボート2連対率,\
5枠_艇番,5枠_登録番号,5枠_選手名,5枠_年齢,5枠_支部,5枠_体重,5枠_級別,\
5枠_全国勝率,5枠_全国2連対率,5枠_当地勝率,5枠_当地2連対率,\
5枠_モーター番号,5枠_モーター2連対率,5枠_ボート番号,5枠_ボート2連対率,\
6枠_艇番,6枠_登録番号,6枠_選手名,6枠_年齢,6枠_支部,6枠_体重,6枠_級別,\
6枠_全国勝率,6枠_全国2連対率,6枠_当地勝率,6枠_当地2連対率,\
6枠_モーター番号,6枠_モーター2連対率,6枠_ボート番号,6枠_ボート2連対率\n"

TABLE_NAME = "boat_race_table" # 格納先のテーブル名

logger = logging.getLogger(__name__)

def rename_columns(df): #fileのカラム名とtableのカラム名が違う場合に、変更
    df.rename(columns={
        "レース日":"race_date",
        "レース場":"race_track",
        "レース回":"race_number",
        "1枠_艇番":"box_number_1",
        "1枠_登録番号":"registration_number_1",
        "1枠_選手名":"player_name_1",
        "1枠_年齢":"age_1",
        "1枠_支部":"branch_1",
        "1枠_体重":"body_weight_1",
        "1枠_級別":"class_1",
        "1枠_全国勝率":"national_win_percentage_1",
        "1枠_全国2連対率":"national_2_way_percentage_1",
        "1枠_当地勝率":"local_win_percentage_1",
        "1枠_当地2連対率":"local_2_way_percentage_1",
        "1枠_モーター番号":"motor_number_1",
        "1枠_モーター2連対率":"motor_2_way_percentage_1",
        "1枠_ボート番号":"boat_number_1",
        "1枠_ボート2連対率":"boat_2_way_percentage_1",
        "2枠_艇番":"box_number_2",
        "2枠_登録番号":"registration_number_2",
        "2枠_選手名":"player_name_2",
        "2枠_年齢":"age_2",
        "2枠_支部":"branch_2",
        "2枠_体重":"body_weight_2",
        "2枠_級別":"class_2",
        "2枠_全国勝率":"national_win_percentage_2",
        "2枠_全国2連対率":"national_2_way_percentage_2",
        "2枠_当地勝率":"local_win_percentage_2",
        "2枠_当地2連対率":"local_2_way_percentage_2",
        "2枠_モーター番号":"motor_number_2",
        "2枠_モーター2連対率":"motor_2_way_percentage_2",
        "2枠_ボート番号":"boat_number_2",
        "2枠_ボート2連対率":"boat_2_way_percentage_2",
        "3枠_艇番":"box_number_3",
        "3枠_登録番号":"registration_number_3",
        "3枠_選手名":"player_name_3",
        "3枠_年齢":"age_3",
        "3枠_支部":"branch_3",
        "3枠_体重":"body_weight_3",
        "3枠_級別":"class_3",
        "3枠_全国勝率":"national_win_percentage_3",
        "3枠_全国2連対率":"national_2_way_percentage_3",
        "3枠_当地勝率":"local_win_percentage_3",
        "3枠_当地2連対率":"local_2_way_percentage_3",
        "3枠_モーター番号":"motor_number_3",
        "3枠_モーター2連対率":"motor_2_way_percentage_3",
        "3枠_ボート番号":"boat_number_3",
        "3枠_ボート2連対率":"boat_2_way_percentage_3",
        "4枠_艇番":"box_number_4",
        "4枠_登録番号":"registration_number_4",
        "4枠_選手名":"player_name_4",
        "4枠_年齢":"age_4",
        "4枠_支部":"branch_4",
        "4枠_体重":"body_weight_4",
        "4枠_級別":"class_4",
        "4枠_全国勝率":"national_win_percentage_4",
        "4枠_全国2連対率":"national_2_way_percentage_4",
        "4枠_当地勝率":"local_win_percentage_4",
        "4枠_当地2連対率":"local_2_way_percentage_4",
        "4枠_モーター番号":"motor_number_4",
        "4枠_モーター2連対率":"motor_2_way_percentage_4",
        "4枠_ボート番号":"boat_number_4",
        "4枠_ボート2連対率":"boat_2_way_percentage_4",
        "5枠_艇番":"box_number_5",
        "5枠_登録番号":"registration_number_5",
        "5枠_選手名":"player_name_5",
        "5枠_年齢":"age_5",
        "5枠_支部":"branch_5",
        "5枠_体重":"body_weight_5",
        "5枠_級別":"class_5",
        "5枠_全国勝率":"national_win_percentage_5",
        "5枠_全国2連対率":"national_2_way_percentage_5",
        "5枠_当地勝率":"local_win_percentage_5",
        "5枠_当地2連対率":"local_2_way_percentage_5",
        "5枠_モーター番号":"motor_number_5",
        "5枠_モーター2連対率":"motor_2_way_percentage_5",
        "5枠_ボート番号":"boat_number_5",
        "5枠_ボート2連対率":"boat_2_way_percentage_5",
        "6枠_艇番":"box_number_6",
        "6枠_登録番号":"registration_number_6",
        "6枠_選手名":"player_name_6",
        "6枠_年齢":"age_6",
        "6枠_支部":"branch_6",
        "6枠_体重":"body_weight_6",
        "6枠_級別":"class_6",
        "6枠_全国勝率":"national_win_percentage_6",
        "6枠_全国2連対率":"national_2_way_percentage_6",
        "6枠_当地勝率":"local_win_percentage_6",
        "6枠_当地2連対率":"local_2_way_percentage_6",
        "6枠_モーター番号":"motor_number_6",
        "6枠_モーター2連対率":"motor_2_way_percentage_6",
        "6枠_ボート番号":"boat_number_6",
        "6枠_ボート2連対率":"boat_2_way_percentage_6",
        "レースコード":"race_id"
    }, inplace=True)
    return df

# テキストファイルからデータを抽出し、CSVファイルに書き込む関数
def get_data(text_file, path):
    # CSVファイルを追記モードで開く
    csv_file = open(path, "a", encoding="shift_jis")

    # テキストファイルからデータを抽出
    for contents in text_file:

        # 文字列「競争成績」を検索してタイトル・日次・レース日・レース場を取得
        # rは正規表現でraw文字列を指定するおまじない
        if re.search(r"番組表", contents):
            # 1行スキップ
            text_file.readline()
            line = text_file.readline()

            # 1行スキップ
            text_file.readline()

            # 日次、レース日、レース場を取得
            line = text_file.readline()
            date = line[17:28]
            stadium = line[52:55]

        # キーワード「電話投票締切予定」を見つけたら
        if re.search(r"電話投票締切予定", contents):

            # キーワードを見つけた行を格納
            line = contents

            # レース名にキーワード「進入固定」が割り込んだ際の補正(「進入固定戦隊」は除くためＨまで含めて置換)
            if re.search(r"進入固定", line):
                line = line.replace('進入固定 Ｈ', '進入固定     Ｈ')

            # レース回
            race_round = line[0:3]

            # 4行スキップ(ヘッダー部分)
            text_file.readline()
            text_file.readline()
            text_file.readline()
            text_file.readline()

            # 選手データを格納する変数を定義
            racer_data = ""

            # 選手データを読み込む行(開始行)を格納
            line = text_file.readline()

            # 空行またはキーワード「END」まで処理を繰り返す = 1～6艇分の選手データを取得
            while line != "\n":

                if re.search(r"END", line):
                    break

                # 選手データを格納(行末にカンマが入らないように先頭にカンマを入れる)
                racer_data += "," + line[0] + "," + line[2:6] + "," + line[6:10] + "," + line[10:12] \
                              + "," + line[12:14] + "," + line[14:16] + "," + line[16:18] \
                              + "," + line[19:23] + "," + line[24:29] + "," + line[30:34] \
                              + "," + line[35:40] + "," + line[41:43] + "," + line[44:49] \
                              + "," + line[50:52] + "," + line[53:58]

                # 次の行を読み込む
                line = text_file.readline()
                
            # 抽出したデータをCSVファイルに書き込む
            csv_file.write(date + "," + stadium + "," + race_round
                        + racer_data + "\n")

    # CSVファイルを閉じる
    csv_file.close()

def customfunc(event):
    logger.info("Begin function customfunc")

    # URL生成
    date = dt.now().today()
    date_str = date.strftime("%Y%m%d")
    yyyymm = date_str[0:4] + date_str[4:6]
    yymmdd = date_str[2:4] + date_str[4:6] + date_str[6:8]

    variable_url = FIXED_URL + yyyymm + "/b" + yymmdd + ".lzh"
    file_name = "b" + yymmdd + ".lzh"

    # bucket接続
    data_store = connect('data_store')
    bucket_connect = data_store.get_bucket('boat-player-bucket')

    # ダウンロード
    bucket_connect.put_request(variable_url, path=file_name)

    # 解凍
    local_filepath = "/tmp/{}".format(file_name)
    bucket_connect.fget(file_name, local_filepath)  # バケットのfput関数でバケットのファイルをtmp領域に移動します
    unzip_file = lhafile.Lhafile(local_filepath)
    info = unzip_file.infolist()
    name = info[0].filename
    local_unzip_filepath = "/tmp/{}.TXT".format(file_name.split(".")[0])
    open(local_unzip_filepath, "wb").write(unzip_file.read(name))

    # TXTファイルをCSVに変換
    local_csv_path = "/tmp/{}.csv".format(file_name.split(".")[0])
    csv_file = open(local_csv_path, "w", encoding="shift_jis")
    csv_file.write(CSV_FILE_HEADER)
    csv_file.close()

    # テキストファイルを開く
    text_file = open(local_unzip_filepath, "r", encoding="shift_jis")
    # データを抽出する
    get_data(text_file, local_csv_path)
    # テキストファイルを閉じる
    text_file.close()
    # CSVファイルをバケットに保存
    bucket_connect.fput("{}.csv".format(file_name.split(".")[0]), local_csv_path)
    
    # テーブルに追加する
    df = pd.read_csv(local_csv_path, encoding="shift-jis")

    # 前処理
    # レース場の空白を削除する
    df["レース場"] = df["レース場"].map(lambda x:x.replace('　',''))

    # 全角数字を半角数字にする
    # 変換テーブルを定義
    trans_asc = str.maketrans('１２３４５６７８９０Ｒ：　', '1234567890R: ')
    # 変換テーブルを用いて文字列を置換
    df["レース回"] = df["レース回"].map(lambda x:x.translate(trans_asc).replace(' ','0'))
    df["レース日"] = df["レース日"].map(lambda x:x.translate(trans_asc).replace(' ','0'))

    # 空白に0を入れる
    df = df.replace(" ", 0)
    # レースコードを追加
    # レース場コードを生成
    dict_stadium = {'桐生': 'KRY', '戸田': 'TDA', '江戸川': 'EDG', '平和島': 'HWJ',
                    '多摩川': 'TMG', '浜名湖': 'HMN', '蒲郡': 'GMG', '常滑': 'TKN',
                    '津': 'TSU', '三国': 'MKN', '琵琶湖': 'BWK', '住之江': 'SME',
                    '尼崎': 'AMG', '鳴門': 'NRT', '丸亀': 'MRG', '児島': 'KJM',
                    '宮島': 'MYJ', '徳山': 'TKY', '下関': 'SMS', '若松': 'WKM',
                    '芦屋': 'ASY', '福岡': 'FKO', '唐津': 'KRT', '大村': 'OMR','びわこ': 'BWK'
                    }
    df["レースコード"] = df["レース日"].map(lambda x:x.split("年")[0]) + df["レース日"].map(lambda x:x.split("年")[1].split("月")[0]) + df["レース日"].map(lambda x:x.split("年")[1].split("月")[1].split("日")[0]) + df["レース場"].map(lambda x:dict_stadium[x]) +df["レース回"].map(lambda x:x[:-1])

    df = rename_columns(df)

    # 選手ごとのレコードに加工
    df_player = pd.DataFrame(columns=[
        "race_id",
        "race_date",
        "race_track",
        "age", 
        "boat_2_way_percentage", 
        "boat_number",
        "body_weight",
        "box_number",
        "branch",
        "class",
        "local_2_way_percentage",
        "local_win_percentage",
        "motor_2_way_percentage",
        "motor_number",
        "national_2_way_percentage",
        "national_win_percentage",
        "player_name",
        "race_number",
        "registration_number",
    ])

    for counter in range(6):
        df_tmp = df[[
            "race_id", 
            "race_date",
            "race_track",
            "age_{}".format(counter+1), 
            "boat_2_way_percentage_{}".format(counter+1), 
            "boat_number_{}".format(counter+1),
            "body_weight_{}".format(counter+1),
            "box_number_{}".format(counter+1),
            "branch_{}".format(counter+1),
            "class_{}".format(counter+1),
            "local_2_way_percentage_{}".format(counter+1),
            "local_win_percentage_{}".format(counter+1),
            "motor_2_way_percentage_{}".format(counter+1),
            "motor_number_{}".format(counter+1),
            "national_2_way_percentage_{}".format(counter+1),
            "national_win_percentage_{}".format(counter+1),
            "player_name_{}".format(counter+1),
            "race_number",
            "registration_number_{}".format(counter+1),
        ]]
        
        df_tmp = df_tmp.rename(columns={
            "age_{}".format(counter+1): "age", 
            "boat_2_way_percentage_{}".format(counter+1):"boat_2_way_percentage", 
            "boat_number_{}".format(counter+1):"boat_number",
            "body_weight_{}".format(counter+1):"body_weight",
            "box_number_{}".format(counter+1):"box_number",
            "branch_{}".format(counter+1):"branch",
            "class_{}".format(counter+1):"class",
            "local_2_way_percentage_{}".format(counter+1):"local_2_way_percentage",
            "local_win_percentage_{}".format(counter+1):"local_win_percentage",
            "motor_2_way_percentage_{}".format(counter+1):"motor_2_way_percentage",
            "motor_number_{}".format(counter+1):"motor_number",
            "national_2_way_percentage_{}".format(counter+1):"national_2_way_percentage",
            "national_win_percentage_{}".format(counter+1):"national_win_percentage",
            "player_name_{}".format(counter+1):"player_name",
            "registration_number_{}".format(counter+1):"registration_number",
        })

        df_player = pd.concat([df_player, df_tmp])

    cn = connect("dwh/data_prim/")
    df_player.rename(columns = str.lower, inplace=True)
    df_rows = df_player.shape[0]
    stats = bulk_insert(cn, TABLE_NAME, df_player)

    # ダウンロードした圧縮ファイル削除
    bucket_connect.delete(file_name)

    # 接続解除
    del data_store
    del bucket_connect
    del cn
    logger.info("END function customfunc")